// for copying and licensing see the attached file LICENSE.txt; please keep this file with the software
//
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>
#include <openssl/x509v3.h>
#include "networking.h"
#include "sslcrypto.h"

int verbosity = 1; 

int setmicroseconds( struct timeval *tv, const char *mstxt, long jiffie ){
  long allms; char *endptr; int ret = 0;
  errno = 0; allms = strtol( mstxt, &endptr, 0 );
  if( errno != 0 || ( endptr != NULL && *endptr != '\000' ) ) return EINVAL;
  if( allms < 0 ) return E2BIG;  // time_t is known to be at least a long
  jiffie = 1000000 / jiffie; if( allms < jiffie ) { allms = jiffie; ret = ERANGE; }
  tv->tv_sec = (time_t) allms / 1000000; 
  tv->tv_usec =  (suseconds_t) allms % 1000000; 
  return ret;
}

#define NanoFraction 1000000000

int setnanoseconds( struct timespec *ts, const char *mstxt, long jiffie ){
  long long allns; char *endptr;
  errno = 0; allns = strtoll( mstxt, &endptr, 0 );
  if( errno != 0 || ( endptr != NULL && *endptr != '\000' ) ) return EINVAL;
  if( allns < 0 || allns / NanoFraction > MaxSigned(time_t) ) return E2BIG;
  ts->tv_sec = (time_t) allns / NanoFraction; 
  ts->tv_nsec =  (long) allns % NanoFraction; 
  return 0;
}

bool checkascii( const char *str ) {
  while(*str) if(*str<' ') return false; else str++;
  return true;
}

int doclose(int fd) { int ret; while( ( ret = close(fd) ) && ret == EINTR ); return ret; }
void dup2quitonerr( int oldfd, int newfd ) { int ret; do { ret = dup2(oldfd,newfd); } while( ret==EINTR || ret==EBUSY ); if(ret!=newfd) { perror("error by dup2 when daemonizing"); exit(2); } }

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

BIO *membio = NULL;
BIO *stdout_bio = NULL;

const EVP_MD *md_sha1;
const EVP_MD *md_sha256;
const EVP_MD *md_sha512;

int netwcomm_init() {
  membio = BIO_new(BIO_s_mem());
  stdout_bio = BIO_new_fp(stdout,BIO_NOCLOSE);
  md_sha1 = EVP_sha1(); md_sha256 = EVP_sha256(); md_sha512 = EVP_sha512();
  if(!membio||!stdout_bio||!md_sha256||!md_sha512) return -1;
  return 0;
}

void netwcomm_cleanup() {
  if(membio) BIO_free(membio);
  if(stdout_bio) BIO_free(stdout_bio);
}

//enum ShaType { ST_SHA256, ST_SHA512 };
//enum ShaWhatOfCert { SK_FULLCERT, SK_PUBKEY };
int shalen[ST_count] = { -1, 32, 64, 20 };

bool sha_of_cert( uchar *digest, int shaWhat, int shaType, X509 *cert, const char *err_comment ) {
  X509_PUBKEY *pubkey; uchar *der_pubkey = NULL; int der_len; unsigned int digest_len; const EVP_MD *md;
  switch(shaWhat) {
    case SK_PUBKEY:
      pubkey = X509_get_X509_PUBKEY(cert); if(!pubkey) return false;
      der_len = i2d_X509_PUBKEY(pubkey,&der_pubkey);
      // X509_PUBKEY_free(pubkey);  // must not be freed: is part of cert; will be freed automatically when cert is freed
      if(der_len>=0) switch(shaType) {
	case ST_SHA1: SHA1( der_pubkey, der_len, digest ); break;
	case ST_SHA256: SHA256( der_pubkey, der_len, digest ); break;
	case ST_SHA512: SHA512( der_pubkey, der_len, digest ); break;
	default: 
	  if(err_comment) fprintf(stderr,"DANE: sha_of_cert: ShaType %i not implemented (%s)\n", shaType, err_comment ); 
	             else fprintf(stderr,"DANE: sha_of_cert: ShaType %i not implemented\n", shaType); 
	  return false;
      }
      OPENSSL_free(der_pubkey);
      return der_len >= 0;
    case SK_FULLCERT:
      switch(shaType) { case ST_SHA1: md = md_sha1; break; case ST_SHA256: md = md_sha256; break; case ST_SHA512: md = md_sha512; break; default: if(err_comment) fprintf(stderr,"DANE: sha_of_cert: ShaType %i not implemented (%s)\n",shaType,err_comment); else fprintf(stderr,"DANE: sha_of_cert: ShaType %i not implemented\n",shaType); return false; }
      return X509_digest(cert, md, digest, &digest_len);
    default: 
      if(err_comment) fprintf(stderr,"DANE: sha_of_cert: unknown ShaWhat mode (%i) - %s\n", shaWhat, err_comment); 
                 else fprintf(stderr,"DANE: sha_of_cert: unknown ShaWhat mode (%i)\n", shaWhat); 
      return false;
  }
  return true;
}

int X509name2utf( BIO *bio, X509_NAME *nm, int indent ) {
  int flags;
  // doku: for utf8 terminals ASN1_STRFLGS_ESC_MSB should be unset
  if( indent >= 0 ) { flags = XN_FLAG_MULTILINE & ~ASN1_STRFLGS_ESC_MSB; }  
               else { flags = XN_FLAG_ONELINE & ~ASN1_STRFLGS_ESC_MSB; indent = 0;}
  // not necessary / appears to be a noop: BIO_reset(membio); BIO_flush(membio);
  return X509_NAME_print_ex( bio, nm, indent, flags );
  // would return length of full BIO buffer, including allocation for previous calls: BIO_number_written( membio );
};

bool print_v3_ext( FILE *out, X509 *cert, int nid, const char *tagname, int indent ) {
  int index; X509_EXTENSION *ext;
  index = X509_get_ext_by_NID(cert,nid,-1);
  if (index >= 0) {
    fprintf(out,tagname);
    ext = X509_get_ext(cert, index);
    X509V3_EXT_print_fp(out, ext, 0, indent);   // flag=0, indent=0
    return true;
  }
  return false;
}

void prn_as_hex(FILE *out, const char *message, const uchar *digest, int len, int flags ) { int i;
  bool prnsep, prnnl; prnsep=!(flags&PRNHEX_NOSEP); prnnl=!(flags&PRNHEX_NONL);
  fprintf(out,message); for( i=0; i<len;i++) fprintf(out,i&&prnsep?":%02X":"%02X",digest[i]); if(prnnl) fputc('\n',out); 
}

#define checkret if(ret<0) fprintf(out,"<failed to print>");

void ShowCert(X509 *cert,const char *whatacert,FILE *out) {
  BIO *out_bio; int ret; uchar digest[64]; EVP_PKEY *pbk; struct pubkey pbk_decoded; BIGNUM *serial; uchar *bnbuf; int buflen; char *decserial; // ASN1_PCTX *pctx; // char *line;
  if ( cert != NULL ) {
    if(whatacert) printf("%s certificate:\n",whatacert);
    out_bio = BIO_new_fp(out,BIO_NOCLOSE);
    //line = X509_NAME_oneline(X509_get_subject_name(cert), 0, 0);
    //fprintf(out," subject: %s\n", line);
    //free(line);       /* free the malloc'ed string */
    fprintf(out," subject:\n"); ret=X509name2utf( out_bio, X509_get_subject_name(cert), 3 ); checkret; fputc('\n',out);
    fprintf(out," issuer:\n"); ret=X509name2utf( out_bio, X509_get_issuer_name(cert), 3 ); checkret; fputc('\n',out);
    fprintf(out," not before: "); ret=ASN1_TIME_print( out_bio,  X509_get0_notBefore(cert)); checkret; fputc('\n',out);
    fprintf(out," not after: "); ret=ASN1_TIME_print( out_bio, X509_get0_notAfter(cert)); checkret; fputc('\n',out);
    serial = ASN1_INTEGER_to_BN(X509_get_serialNumber(cert),NULL);
    bnbuf = alloca(buflen=BN_num_bytes(serial));
    BN_bn2bin(serial,bnbuf);
    prn_as_hex(out," serial: ", bnbuf,buflen,PRNHEX_NONL);
    decserial = BN_bn2dec(serial); printf(", %s\n", decserial); OPENSSL_free(decserial);
    BN_free(serial);
    //fprintf(out," CN~common name, C~country, O~organization, OU~organization unit/url, L~locality/city, ST~state or province\n");
    if( print_v3_ext(out, cert, NID_subject_alt_name, " " LN_subject_alt_name ": ",0) ) fputc('\n',out);
    print_v3_ext(out, cert, NID_info_access, " " LN_info_access ":\n",3);
    //pctx = ASN1_PCTX_new(); ASN1_PCTX_set_flags(pctx,ASN1_PCTX_FLAGS_SHOW_TYPE);
    //fprintf(out," public key: "); ret = EVP_PKEY_print_params( out_bio, X509_PUBKEY_get0(X509_get_X509_PUBKEY(cert)), 3, pctx ); checkret; fputc('\n',out);
    //ASN1_PCTX_free(pctx);
    pbk = X509_PUBKEY_get0(X509_get_X509_PUBKEY(cert));
    if( parse_pubkey(&pbk_decoded,pbk) ) {
      ASN1_OBJECT *obj; char *nid2txt_buf; int len;
      obj = OBJ_nid2obj(pbk_decoded.type_nid);
      len = OBJ_obj2txt(NULL,0,obj,0)+1; nid2txt_buf=(char*)alloca(len);
      OBJ_obj2txt(nid2txt_buf, len, obj, 0);
      fprintf(out," key type: %s\n",nid2txt_buf);
      if( pbk_decoded.flags & PBK_KNOW_KEYSIZE ) fprintf(out," key size: %ibit\n",get_keysize(&pbk_decoded));
      free_pubkey(&pbk_decoded);
    } else fprintf(out," key type: unknown\n");
    if(!(pbk_decoded.flags&PBK_KNOW_KEYSIZE)) fprintf(out," key size: ~%ibit\n",8*i2d_PUBKEY(pbk,NULL));
    ret = sha_of_cert( digest, SK_FULLCERT, ST_SHA1, cert, NULL );
    if(ret) prn_as_hex(out," sha1(full-cert) = ",digest,shalen[ST_SHA1],0); else fprintf(out,"sha256(full-cert): <failed to print>\n");
    ret = sha_of_cert( digest, SK_FULLCERT, ST_SHA256, cert, NULL );
    if(ret) prn_as_hex(out," sha256(full-cert) = ",digest,shalen[ST_SHA256],0); else fprintf(out,"sha256(full-cert): <failed to print>\n");
    ret = sha_of_cert( digest, SK_FULLCERT, ST_SHA512, cert, NULL );
    if(ret) prn_as_hex(out," sha512(full-cert) = ",digest,shalen[ST_SHA512],0); else fprintf(out,"sha512(full-cert): <failed to print>\n");
    ret = sha_of_cert( digest, SK_PUBKEY, ST_SHA256, cert, NULL );
    if(ret) prn_as_hex(out," sha256(pubkey) = ",digest,shalen[ST_SHA256],0); else fprintf(out,"sha256(pubkey): <failed to print>\n");
    ret = sha_of_cert( digest, SK_PUBKEY, ST_SHA512, cert, NULL );
    if(ret) prn_as_hex(out," sha512(pubkey) = ",digest,shalen[ST_SHA512],0); else fprintf(out,"sha512(pubkey): <failed to print>\n");
    fputc('\n',out);
    if(out_bio) BIO_free(out_bio);
  }
  else
      printf("info: No %s%scertificate configured.\n",whatacert?whatacert:"",whatacert?" ":"");
}

#undef checkret

bool CheckCertDate( X509 *cert, const char *msghdr, FILE *out ) {
  const ASN1_TIME *not_before, *not_after; ASN1_TIME *now; time_t time_now; int pday, psec, ret; bool result=true;
  time( &time_now ); now = ASN1_UTCTIME_set( NULL, time_now );
  not_before = X509_get0_notBefore(cert); not_after = X509_get0_notAfter(cert);  // internal pointers; must not be freed
  ret = ASN1_TIME_diff( &pday, &psec, not_before, now ); if(!ret) { fprintf(out,"%sASN1_TIME_diff error: possibly invalid date\n",msghdr); result=false; }
  if( pday < 0 || ( pday == 0 && psec < 0 ) ) { 
    if(out) fprintf(out,"%stimespan error: current time is %i days and %i secs before the cert was issued.\n", msghdr, -pday, -psec );
    result=false;
  } else {
    ret = ASN1_TIME_diff( &pday, &psec, now, not_after );
    if(!ret) { fprintf(out,"%sASN1_TIME_diff error: possibly invalid date\n",msghdr); result=false; }
    else if( pday < 0 || ( pday == 0 && psec < 0 ) ) { 
      if(out) fprintf(out,"%stimespan error: current time is %i days and %i secs after the cert was invalidated.\n", msghdr, -pday, -psec );
      result=false;
    } //else prnbuf_printf(" date is ok.\n");
  }
  ASN1_STRING_free(now);
  return result;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

X509* read_cert_from_file( const char *filename ) {
  int result; X509 *cert = NULL; BIO *bo = NULL;
  bo = BIO_new(BIO_s_file());
  result = BIO_read_filename(bo, filename );
  if(!result) { BIO_free(bo); fprintf(stderr,"error reading certfile '%s' - will refuse to download over https.\n",filename); return NULL; }
  PEM_read_bio_X509(bo, &cert, 0, NULL);
  BIO_free(bo);
  return cert;
}

bool X509_compare_equal( X509 *certA, X509 *certB ) {
  int lenA, lenB; unsigned char *bufA = NULL, *bufB = NULL; bool result;
  lenA = i2d_X509(certA,&bufA); lenB = i2d_X509(certB,&bufB);
  result = ( lenA == lenB ) && ( lenA > 0 ) && !memcmp(bufA,bufB,lenA);
  OPENSSL_free(bufA); OPENSSL_free(bufB);
  return result;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

int directWrite( void *fd, const void *buf, int len ) { return write(*(int*)fd,buf,len); }
int directRead( void *fd, void *buf, int len ) { return read(*(int*)fd,buf,len); }

int sslWrite( void *fd, const void *buf, int len ) { 
  int res = SSL_write((SSL*)fd,buf,len); 
  if(res<=0) {
    res = SSL_get_error((SSL*)fd,res);
    if( res==SSL_ERROR_ZERO_RETURN || res==SSL_ERROR_NONE ) return 0;
    return -1;
  } else return res;
}

uint64 pn = 0;

int sslRead( void *fd, void *buf, int len ) { int res;
  //if(pn++%60==0) return -1;
  //if(pn%60==0) return -1;
  //if(++pn%60==0) sleep(4);
  //usleep(57000);
  res = SSL_read((SSL*)fd,buf,len); 
  if(res<=0) {
    res = SSL_get_error((SSL*)fd,res);
    if( res==SSL_ERROR_ZERO_RETURN || res==SSL_ERROR_NONE ) return 0;
    return -1;
  } else return res;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define cin(c,a,b) ( (a)<=(c) && (c)<=(b) )

int parse_ipv6( char *start, char **host, struct in6_addr *in6addr, uint16_t *port, int *out_k ) {
  int i, k; int result = 1, res; long portnum; char *stopchar;
  if(*start=='[') {
    i = 1+strcspn(start+1,"/]");
    k = i + strcspn(start+i,"/"); *out_k = k;
    if(i==k) {
      fprintf(stderr,"host spec contains opening \"[\" but no closing \"]\".\n");
      result = 2;
    } else {
      // start[i]==']'
      if(start[i+1]==':') {
  parse_port_number:
	errno = 0; portnum = strtol( start+(i+2), &stopchar, 0 );
	if( errno!=0 || ( *stopchar && *stopchar != '/' && *stopchar != '\000' ) || portnum<=0 || portnum>=65535 ) {
	  fprintf(stderr,"error: '"); fwrite( start+(i+2), k-(i+2), 1, stderr); fprintf(stderr,"' is not a valid port number.\n\n");
	  *host = NULL;
	  return -1;
	} else *port = (uint16_t) portnum;
      } else if(k!=i+1) { fprintf(stderr,"stale characters after \"]\" in ipv6 spec\n"); result = 2; }
    }
    *host = malloc(i); memcpy(*host,start+1,i-1); (*host)[i-1]='\000';
  } else {
    bool mayip6 = false; register char c;
    i=0; while( start[i] && start[i]!='/' && start[i]!=']' ) {
      c = start[i++];
      if(!( cin(c,'0','9') || cin(c,'a','f') || cin(c,'A','F') || c==':' || c=='.' )) { *host=NULL; return 0; }
      if(c==':') mayip6 = true;
      else if( c=='.' && !mayip6 ) { *host=NULL; return 0; }
    }
    if(!mayip6) { *host=NULL; return 0; }
    if(start[i]==']') { 
      fprintf(stderr,"stale \"]\" in ipv6 spec discovered.\n"); 
      k = i+1 + strcspn(start+(i+1),"/"); result = 2; 
    } else k=i;
    *out_k = k;
    if( start[i]==']' && start[i+1]==':' ) { start--; i++; goto parse_port_number; }
    *host = malloc(i+1); memcpy(*host,start,i); (*host)[i]='\000';
  }
  res = inet_pton(AF_INET6,*host,in6addr);
  if(res!=1) { 
    fprintf(stderr,"not a correct ipv6 addr: \"%s\".\n\n",*host);
    free(*host); *host=NULL; 
    return -1; 
  }
  return result;
}

int parse_urlspec( char *url, struct UrlSpec *urlspec, bool force_ipv6 ) { 
  int i, j, k; long portnum; char *stopchar; int gai_errno; uint16_t port;
  uint16_t default_port = 80; int check4ip6adr; struct in6_addr in6addr;
  i = strcspn(url,"/\\");
  if( i <= 0 || url[i-1] != ':' || url[i] != '/' || url[i+1] != '/' ) {
    fprintf(stderr,"expected protocol://path: %s\n",url);
    return 2;
  } else if ( i-1<4 || 5<i-1 || memcmp(url,"http",4) || ( i-1==5 && url[4]!='s' ) ) {
    fprintf(stderr,"unknown protocol: "); fwrite(url,i-1,1,stderr); fputc('\n',stderr);
    return 3;
  }
  urlspec->fullURL = url;
  switch(i-1) {
    case 4: urlspec->protocol = PR_HTTP; default_port = 80; break;
    case 5: urlspec->protocol = PR_HTTPS; default_port = 443; break;
    default: fprintf(stderr,"protocol (len=%i) not implemented\n",i); return 1;
  }
  i += 2; port = default_port;
  check4ip6adr = parse_ipv6( url+i, &urlspec->host, &in6addr, &port, &k );
  if(check4ip6adr==-1) return 2;
  urlspec->isip6adr = check4ip6adr > 0;
  if(check4ip6adr) k+=i;
  else {
    k = i + strcspn(url+i,"/");
    j = k - 1; while( url[j]!=':' && url[j]!='/' ) j--;
    if( url[j]==':' ) {
      errno = 0; portnum = strtol( url+j+1, &stopchar, 0 );
      if( errno!=0 || ( *stopchar && *stopchar != '/' && *stopchar != '\000' ) || portnum<=0 || portnum>=65535 ) {
	fprintf(stderr,"error: '"); fwrite( url+(j+1), k-(j+1), 1, stderr); fprintf(stderr,"' is not a valid port number.\n\n");
	return 1;
      } else port = (uint16_t) portnum;
    } else {
      j = k;
    }
    urlspec->host = malloc(j-i+1);  // i is one char before start: accounts for len
    memcpy( urlspec->host, url+i, j-i);
    ((char*)urlspec->host)[j-i] = '\000';
  }
  if( port == default_port ) urlspec->port_override = -1;
  else urlspec->port_override = port;
  if( url[k]=='\000' || !strcmp(url+k,"/") ) {
    urlspec->path = "/";
    urlspec->target = NULL;
  } else {
    urlspec->path = url + k;
    urlspec->target = strrchr(url+k,'/'); if(urlspec->target) { urlspec->target++; if(*urlspec->target=='\000') urlspec->target=NULL; }
  }
  struct addrinfo template;
  memset( &template, 0, sizeof(template) ); 
  template.ai_flags = AI_V4MAPPED | AI_ALL;  // no IPv6 found: return IPv4-mapped IPv6 addresses  +AI_ALL return both IPv6 and IPv4-mapped IPv6 addresses
  template.ai_family = force_ipv6 ? AF_INET6 : AF_UNSPEC; template.ai_socktype = SOCK_STREAM; template.ai_protocol = 0;  //  that would be for tcp but we will not care about this value later on
  urlspec->addr = NULL; gai_errno = getaddrinfo( urlspec->host, NULL, &template, &urlspec->addr );
  if(gai_errno) { fprintf(stderr,"getaddrinfo(%s): %s.\n\n",urlspec->host, gai_strerror(gai_errno) ); return 2; }
  if(!urlspec->addr) { fprintf(stderr,"getaddrinfo(%s) returned no result.\n\n",urlspec->host); return 2; }
  struct sockaddr *adr = urlspec->addr->ai_addr;
  if(adr->sa_family==AF_INET) ((struct sockaddr_in*)adr)->sin_port = htons( port );
  else if(adr->sa_family==AF_INET6) ((struct sockaddr_in6*)adr)->sin6_port = htons( port );
  else { fprintf(stderr,"unsupported protocol (neither IPv4 nor IPv6)\n"); return 2; }
  return 0;
}

void print_sockaddr( FILE* out, struct sockaddr *sa ) {
  if( sa->sa_family == AF_INET ) {
    struct sockaddr_in* sa4 = (struct sockaddr_in*)sa;
    long inaddr = ntohl(sa4->sin_addr.s_addr); int i;
    if( inaddr == INADDR_ANY ) fprintf(out,"*");
    else for(i=3;i>=0;i--) { fprintf(out,"%u", (unsigned int) inaddr >> 24 ); inaddr <<= 8; if(i) fputc('.',out); }
    fprintf(out,":%i",ntohs(sa4->sin_port));
  } else if( sa->sa_family == AF_INET6 ) {
    struct sockaddr_in6 *sa6 = (struct sockaddr_in6*)sa; char adrbuf[INET6_ADDRSTRLEN]; const char *adrstr;
    fputc('[',out);
    adrstr = inet_ntop(sa->sa_family,&sa6->sin6_addr,adrbuf,INET6_ADDRSTRLEN);
    if(adrstr) fputs(adrstr,out); else fputc('?',out);
    fprintf(out,"]:%i",ntohs(sa6->sin6_port));
  } else fprintf(out,"unknown/non-IP address family");
}

const char* protocol2str[3] = { NULL, "http", "https" };

void print_urlspec( FILE *out, struct UrlSpec *urlspec ) {
  fprintf(out,"protocol: %s\n",protocol2str[urlspec->protocol]);
  fprintf(out,"addr: "); print_sockaddr(out,urlspec->addr->ai_addr); fputc('\n',out);
  fprintf(out,"host: %s\n", urlspec->host);
  fprintf(out,"path: %s\n", urlspec->path);
  fprintf(out,"target: %s\n\n", urlspec->target ? urlspec->target : "NULL");
}


int socket_connect( int *s, struct addrinfo *target_host_ais, int flags ) {
  struct addrinfo *hst; int saved_errno = ESRCH; bool silent = ( flags & SF_SILENT )!=0;
  assert(!( flags & SF_ANYHOST ));

  for( hst = target_host_ais; hst; hst = hst->ai_next ) {
    if( connect( *s, hst->ai_addr, hst->ai_addrlen ) == 0 ) break;
    saved_errno = errno; 
    if( !silent ) perror("error trying to connect tcp socket to address"); 
    errno = saved_errno;
  }

  if ( *s <= -2 ) saved_errno = EAGAIN;    // EXDEV ~ improper link
  // *s == -1: shield for a timeout signal handler which could have set s=-2 or any other value below zero
  if( !hst ) return saved_errno;

  return 0;
}


